#include <string>

#include "Halide.h"
#include "halide_image_io.h"

#define	COLOURS	4 /* RGBA */
#define	BYTES	1 /* 8-bit */

class Pyramid {
	Halide::Func *gpyramid_, *lpyramid_;
	int levels_;
 public:
	Pyramid(int);
	~Pyramid();
	Halide::Func compute(Halide::Func, int);
};

Halide::Func
blur(Halide::Func u16in)
{
	Halide::Var x("x"), y("y"), c("c"), xi, yi;
	Halide::Func blurx("blurx"), blury("blury");

	blurx(x, y, c) =
	    (u16in(x - 1, y, c) + 2 * u16in(x, y, c) + u16in(x + 1, y, c)) / 4;
	blury(x, y, c) =
	    (blurx(x, y - 1, c) + 2 * blurx(x, y, c) + blurx(x, y + 1, c)) / 4;

	blurx.reorder(c, x, y).unroll(c, COLOURS);
	blury.reorder(c, x, y);

	blury.tile(x, y, xi, yi, 256, 32).vectorize(xi, 8).parallel(y);
	blurx.compute_at(blury, x).vectorize(x, 8);
	return blury;
}

Halide::Func
doubleit(Halide::Func u16in)
{
	Halide::Var x("x"), y("y"), c("c");

	Halide::Func resiz("doubleit");
	resiz(x, y, c) = u16in(x / 2, y / 2, c);

	return resiz;
}

Halide::Func
halveit(Halide::Func u16in)
{
	Halide::Var x("x"), y("y"), c("c");

	Halide::Func resiz("halveit");
	resiz(x, y, c) = u16in(x * 2, y * 2, c);
	resiz.reorder(c, x, y);

	return resiz;
}

Pyramid::Pyramid(int levels) : levels_(levels)
{
	gpyramid_ = new Halide::Func[levels + 1];
	lpyramid_ = new Halide::Func[levels + 1];
}

Pyramid::~Pyramid(void)
{
	delete [] gpyramid_;
	delete [] lpyramid_;
}

Halide::Func
Pyramid::compute(Halide::Func u16in, int levels)
{
	assert(levels > 2 && levels <= levels_);

	Halide::Var x("x"), y("y"), c("c");

	gpyramid_[0] = u16in;
	for (int i = 1; i < levels; ++i) {
		auto name = std::string("gpyramid_" + std::to_string(i));
		gpyramid_[i] = Halide::Func(name);
		gpyramid_[i](x, y, c) = halveit(blur(gpyramid_[i-1]).compute_root())(x, y, c);
		gpyramid_[i].reorder(c, x, y).compute_root();
		gpyramid_[i].debug_to_file(name + ".tif");
	}

	lpyramid_[levels-1](x, y, c) = gpyramid_[levels-1](x, y, c);
	for (int i = levels - 2; i >= 0; --i) {
		auto name = std::string("lpyramid_" + std::to_string(i));
		lpyramid_[i] = Halide::Func(name);
		lpyramid_[i](x, y, c) = Halide::cast<uint16_t>(
		    gpyramid_[i](x, y, c) - doubleit(gpyramid_[i+1])(x, y, c));
		lpyramid_[i](x, y, 3) = Halide::cast<uint16_t>(gpyramid_[i](x, y, 3));
		lpyramid_[i].reorder(c, x, y).compute_root();
		lpyramid_[i].debug_to_file(name + ".tif");
		//lpyramid_[i].trace_stores();
	}
	Halide::Func ret = lpyramid_[1];
	return ret;

}

int
main(void)
{
	Halide::Var x("x"), y("y"), c("c");

	Halide::ImageParam input(Halide::type_of<u_int8_t>(), 3, "input");
	input.set_stride(0, COLOURS).set_stride(2, BYTES);

	Halide::Expr clampx = Halide::clamp(x, 0, input.width() - 1);
	Halide::Expr clampy = Halide::clamp(y, 0, input.height() - 1);

	Halide::Func clamped("clamped");
	clamped(x, y, c) = input(clampx, clampy, c);

	Halide::Func u16in("u16in");
	u16in(x, y, c) = Halide::cast<u_int16_t>(clamped(x, y, c));

	Halide::Func output("output"), p("pyramid");

	Pyramid pyr(3);
	p = pyr.compute(u16in, 3);
	output(x, y, c) = Halide::cast<u_int8_t>(p(x, y, c) + 128);
	output(x, y, 3) = Halide::cast<u_int8_t>(255);
	output.output_buffer().set_stride(0, COLOURS).set_stride(2, BYTES);
	output.reorder(c, x, y);
	output.debug_to_file("dbg.tif");
	//output.trace_stores();
	output.print_loop_nest();

	std::vector<Halide::Argument> args = {input};
	output.compile_to_lowered_stmt("dbg.html", args, Halide::HTML);
	output.compile_to_static_library("run_pyramid", args);

	return (0);
}
